/**
 * A Mega324P Temperature Controlled PWM Fan Looped (Non-Interrupt) Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */

 /**
 * Mega164P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2A/PCINT31)
 *							-----------
 */

/* Includes */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>
#include <stdio.h>

/* Global/System Constants */
#define TRUE 	1
#define FALSE 	0
#define CPU_FREQUENCY_HZ	20000000	// CPU Clock Frequency (in Hz)

/* Macros */
#define sbi(data,bit)			(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 			(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)			(data & (1<<bit))	// Get Bit
#define tbi(data,bit)			(data ^= (1<<bit))	// Toggle Bit

/* Constants for Timer0 */
#define TIMER_RES_US	50

#define TIMER_LATENCY_US			10000
#define TIMER_LATENCY_THRESHOLD		((uint16_t) ((TIMER_LATENCY_US / TIMER_RES_US) + 0.5))

/* Constants for PWM Fan */
#define PWM_PULSE_WIDTH_OFF			0
#define	PWM_PULSE_WIDTH_MIN_ON		60
#define PWM_PULSE_WIDTH_MAX_ON		255
#define	PWM_FAN_LATENCY_US			500000
#define PWM_FAN_LATENCY_THRESHOLD	(PWM_FAN_LATENCY_US / TIMER_RES_US)



/* Constants for Temperature Sensor */
#define ADC_REFERENCE_VOLTAGE		5
#define ADC_REFERENCE_RANGE			1024
#define ADC_TEMP_MAX_DEGREE_F		60


// Analog Voltage = Actual Temp * Sensor Voltage Per Degree * Op Amp Scale = Actual Temp * 20 mV
//		77 F * 20 mV => 1.54 V
//		100 F * 20 mV => 2 V
//		150 F * 20 mV => 3 V
// ADC Value = Analog Voltage / ADC Reference * ADC Range = Analog Voltage / 5 * 1024
//		1.54 V / 5 V * 1024 => 315.392
//		2 V / 5 V * 1024 => 409.6
//		3 V / 5 V * 1024 => 614.4
// Actual Temp = ADC Value / ((Sensor Voltage Per Degree * Op Amp Scale) / (ADC_REFERENCE_VOLTAGE * 1000)) = ADC Value / 4.096
//				= ADC >> 2
//		315.392 / 4.096 = 77 F
//		315.392 >> 2 = 78.848 F

/* Global Variables */
volatile uint32_t pwm_fan_latency_tick_count = 0;
volatile uint16_t timer_tick_count = 0;

/* Prototype Functions */
void init_mcu(void);
void init_adc(void);
void init_timer2_pwm(void);
void init_timer0(void);

/**
 * Triggered every tick (depends on tick_res_us)
 */
ISR(TIMER0_COMPA_vect) {
	pwm_fan_latency_tick_count++;	
	timer_tick_count++;
}

int main(void) {
	static uint8_t adc_data = 0;
	static uint16_t pulse_width = PWM_PULSE_WIDTH_OFF;

	DDRB = 0xFF;
	PORTB = 0xFF;
	
	init_mcu();			// Initialize MCU features
	init_adc();			// Initialize ADC
	init_timer0();
	init_timer2_pwm();
	
	sei();				// Enable interrupts

	_delay_ms(100);		// Delay to stablize initilization

	cbi(PORTB, 0);
	
	while(1) {
		if (pwm_fan_latency_tick_count > PWM_FAN_LATENCY_THRESHOLD) {
			if (gbi(ADCSRA, ADIF)){
				adc_data = ADCH;
				PORTB = ~adc_data;   
				if (adc_data >= ADC_TEMP_MAX_DEGREE_F) {
					pulse_width = PWM_PULSE_WIDTH_MIN_ON + adc_data - ADC_TEMP_MAX_DEGREE_F;
					if (pulse_width > PWM_PULSE_WIDTH_MAX_ON) {
						pulse_width = PWM_PULSE_WIDTH_MAX_ON;
					}
					OCR2A = pulse_width;
				}
				else {
					OCR2A = PWM_PULSE_WIDTH_OFF;
				}
				sbi(ADCSRA, ADIF);
			}
			pwm_fan_latency_tick_count = 0;
		}
	}
	
	return 0;
}

void init_mcu(void) {
	SMCR = 0b00000000;	// 7:4 - Reserved bits
						// 3:1 - Sleep Mode
						//		000 - Idle
						//		001 - ADC Noise Reduction
						//		010 - Power-down
						//		011 - Power-save
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Standby
						//		111 - Extended Standby
						// 0 - Sleep Enable
						
	MCUCR |= (0 << 6);	// 6 - BOD Sleep
	MCUCR |= (0 << 5);	// 5 - BOD Sleep Enable
	
	//PRR = 0b11010110;	// 7 - Power Reduction TWI
						// 6 - Power Reduction Timer2
						// 5 - Power Reduction Timer0
						// 4 - Power Reduction UART1
						// 3 - Power Reduction Timer1
						// 2 - Power Reduction SPI
						// 1 - Power Reduction UART0
						// 0 - Power Reduction ADC
}

void init_adc(void) {
	sbi(DDRA,0);	// PORTA.0 is an input for ADC0
    
	ADMUX = 0b00100000;	// 7:6 - Reference Selection Bits
						//		00 - AREF, internal Vref off
						//		01 - AVCC with external capacitor at AREF pin
						//		10 - Internal 1.1 V Ref with ext cap at AREF
						//		11 - Internal 2.56 V Ref "
						// 5 - ADC Left Adjust Result
						// 4:0 - Analog Channel and Gain Selection Bits
						//		00000 - Single Ended, ADC0
						//		00001 - ", ADC1
						//		00010 - ", ADC2
						//		00011 - ", ADC3
						//		00100 - ", ADC4
						//		00101 - ", ADC5
						//		00110 - ", ADC6
						//		00111 - ", ADC7
						//		01000 - Positive Diferential, ADC0, ADC0, 10x
						//		01001 - ", ADC1, ADC0, 10x
						//		01010 - ", ADC0, ADC0, 200x
						//		01011 - ", ADC1, ADC0, 200x
						//		01100 - ", ADC2, ADC2, 10x
						//		01101 - ", ADC3, ADC2, 10x
						//		01110 - ", ADC2, ADC2, 200x
						//		01111 - ", ADC3, ADC2, 200x
						//		10000 - ", ADC0, ADC1, 1x
						//		10001 - ", ADC1, ADC1, 1x
						//		10010 - ", ADC2, ADC1, 1x
						//		10011 - ", ADC3, ADC1, 1x
						//		10100 - ", ADC4, ADC1, 1x
						//		10101 - ", ADC5, ADC1, 1x
						//		10110 - ", ADC6, ADC1, 1x
						//		10111 - ", ADC7, ADC1, 1x
						//		11000 - ", ADC0, ADC2, 1x
						//		11001 - ", ADC1, ADC2, 1x
						//		11010 - ", ADC2, ADC2, 1x
						//		11011 - ", ADC3, ADC2, 1x
						//		11100 - ", ADC4, ADC2, 1x
						//		11101 - ", ADC5, ADC2, 1x
						//		11110 - Singled Ended Input, 1.1 V (VBG)
						//		11111 - ", 0 V (GND)
	
	ADCSRA = 0b11100111;// 7 - ADC Enable
						// 6 - ADC Start Conversion
						//	=> Single Conversion => Start each conversion
						//	=> Free Running => Start first conversion
						// 5 - ADC Auto Trigger Enable
						// 4 - ADC Interrupt Flag
						// 3 - ADC Interrupt Enable
						// 2:0 - ADC Prescaler Select Bits
						//		000 - 2 => 62.5 ns * 2 = 125 ns
						//		001 - 2 => 125 ns
						//		010 - 4 => 250 ns
						//		011 - 8 => 500 ns
						//		100 - 16 => 1 us
						//		101 - 32 => 2 us
						//		110 - 64 => 4 us
						//		111 - 128 => 8 us
	
	ADCSRB = 0b00000000;// 7:3 - Reserved Bits (6 is used elsewhere)
						// 2:0 - ADC Auto Trigger Source
						//		000 - Free Running Mode
						//		001 - Analog Comparator
						//		010 - External Interrupt Request 0
						//		011 - Timer0 Compare Match
						//		100 - Timer0 Overflow
						//		101 - Timer1 Compare Match B
						//		110 - Timer1 Overlow
						//		111 - Timer1 Capture Event
						
	DIDR0 = 0b00000001;	// 7:0 - ADC7:0 Digital Input Disable
						//	=> Disable digital input buffer to reduce power consumption 
}

/**
 * Enable timer0 to act as a timer for measuring the time
 * between rising and falling and then rising to rising.
 */
void init_timer0(void) {
	// A single cycle = 1/20 MHz = 50 ns

	TCCR0A = 0b00000010;	// 7:6 - Compare Match Output A Mode
							// 5:4 - " B "
							// 3:2 - Reserved
							// 1:0 - Waveform Generation Mode
							//	TCCR0B.3|TCCR0A.1|TCCR0A.0
							//	000 - Normal
							//	010 - CTC
	 
	TCCR0B = 0b00000010;	// 7 - Force Output Compare A
							// 6 - " B
							// 5:4 - Reserved bits
							// 3 - Waveform Generation Mode (see above)
							// 2:0 - Clock select
							//	000 - Timer stopped
							//	001 - 1
							//	010 - 8
							//	011 - 64
							//	100 - 256
							//	101 - 1024
							//	110 - Ext clock source at T0 pin, falling
							//	111 - ", rising
							
	OCR0A = TIMER_RES_US * 2.5;	// OCR0A = timer resolution in us / 400 ns
								// => timer resolution in us * 2.5
	
	TIMSK0 = 0b00000010;	// 7:3 - Reserved bits for other timer interrupts
							// 2 - Timer0 output compare match B interrupt
							// 1 - " A "
							// 0 - Timer0 overflow interrupt
}


void init_timer2_pwm(void) {
	sbi(DDRD, 7);	// Set PORTD.5 as an output for OC1A 
    
	TCCR2A = 0b10000011;	// 7:6 - COM2A1:0: Compare Match Output A Mode
							//		Fast PWM Mode
							//		00 - Disconnected
							//		01 - Toggle
							//		10 - Non-inverting
							//		11 - Inverting 
							// 5:4 - COM2B1:0: " B "
							// 3:2 - Reserved
							// 1:0 - WGM21:0: Waveform Generation Mode
							//		WGM2|WGM1|WGM0
							//		000 - Normal, TOP = 0xFF
							//		001 - PWM, Phase Correct, TOP=0xFF
							//		010 - CTC, TOP = OCRA
							//		011 - Fast PWM, TOP = 0xFF
							//		100 - Reserved
							//		101 - PWM, Phase Correct, TOP = OCRA
							//		110 - Reserved
							//		111 - Fast PWM, TOP = OCRA
	
	TCCR2B = 0b00000111;	// 7 - FOC2A: Force Output Compare A
							// 6 - FOC2B: " B
							// 5:4 - Reserved
							// 3 - WGM22: Waveform Generation Mode
							//		(See TCCR2A)
							// 2:0 - CS22:0: Clock Select
							//		000 - 0
							//		001 - 1
							//		010 - 8
							//		011 - 32
							//		100 - 64
							//		101 - 128
							//		110 - 256
							//		111 - 1024
							
	OCR2A = PWM_PULSE_WIDTH_OFF;
	
	TIMSK2 = 0b00000000;	// 7:3 - Reserved
							// 2 - OCIE2B: Timer2 Output Compare Match B Interrupt Enable
							// 1 - OCIE2A: " A "
							// 0 - TOIE2: Timer2 Overlow Interrupt Enable
}	
