/**
 * A Mega644P Clock Generator w/ Internal Oscillator Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

/* Global Constants */
#define CPU_FREQUENCY_HZ	8000000
  
/* Constants for Clock Generation */
#define CLOCK_FREQUENCY		153600 

/* Constants for Timer1 */
#define TIMER1_DIVIDER		1
#define TIMER1_COMPARE		((uint16_t) (CPU_FREQUENCY_HZ / (TIMER1_DIVIDER * CLOCK_FREQUENCY * 2) + 0.5))

/* Macro Functions */
#define sbi(data,bit)			(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 			(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)			(data & (1<<bit))	// Get Bit
#define tbi(data,bit)			(data ^= (1<<bit))	// Toggle Bit

void init_timer1(void);

/**
 * Main program
 */
int main(void) {
	sbi(DDRD, 5);
	
	init_timer1();		// Initialize Timer
	
	_delay_ms(100);		// Delay to stablize initilization
	
	// Main program loop
	while(1) {
	}
	
	return 0;
}


void init_timer1(void) {
	TCCR1A = 0b01000000;	// 7:6  Compare Output Mode for Channel A
							// 5:4 - Compare Output Mode for Channel B
							// 3:2 - Reserved
							// 1:0 - Waveform Generation Mode
							
	TCCR1B = 0b00001001;	// 7 - Input Capture Noise Canceler
							// 6 - Input Capture Edge Select
							// 5 - Reserved
							// 4:3 - Waveform Generation Mode
							// 2:0 - Clock Select
							
	TCCR1C = 0b00000000;	// 7 - Force Output Compare for Channel A
							// 6 - Force Output Compare for Channel B
							// 5:0 - Reserved
							
	OCR1AH = (TIMER1_COMPARE) >> 8;
	OCR1AL = (TIMER1_COMPARE << 8) >> 8;
	
	TIMSK1 = 0b00000000;	// 7:6 - Reserved
							// 5 - Timer/Counter1, Input Capture Interrupt Enable
							// 4:3 - Reserved
							// 2 - Timer/Counter1, Output Compare B Match Interrupt Enable
							// 1 - Timer/Counter1, Output Compare A Match Interrupt Enable
							// 0 - Timer/Counter1, Overflow Interrupt Enable
	
}
