/**
 * A Mega644P UART TX Timer2 based ADC w/ Internal Oscillator Alternative Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

/* Includes */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>
#include <stdio.h>

/* Global/System Constants */
#define TRUE	1
#define FALSE	0

#define sbi(data,bit)			(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 			(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)			(data & (1<<bit))	// Get Bit
#define tbi(data,bit)			(data ^= (1<<bit))	// Toggle Bit

/* Timer2/Sleep Constants */
#define TIMER_RES_MS		32  

/* Constants for Timer */
#define TIMER_LATENCY_MS	200
#define TIMER_LATENCY_THRESHOLD		((uint16_t) ((TIMER_LATENCY_MS / TIMER_RES_MS) + 0.5))

/* ADC Constants */
#define ADC_DATA_LENGTH		3

/* Global Variables */
volatile uint8_t timer2_tick_count;
uint8_t adc_data[ADC_DATA_LENGTH], adc_index, adc_flag;

/* Prototype Functions */
void init_mcu(void);
void init_uart0_tx(void);
void init_timer2(void);
void init_adc(void);
void adc_set_channel(uint8_t chan);
void uart0_send_byte_array(uint8_t *data, uint8_t length);
void uart0_send_byte(uint8_t data);

ISR (TIMER2_COMPA_vect) {
	timer2_tick_count++;
}    


/**
 * UART transmission function for byte arrays
 *
 * Can be used to transmitt strings
 * Length terminated (versus null terminated)
 */
void uart0_send_byte_array(uint8_t *data, uint8_t length) {
	for (uint8_t i=0; i<length; i++) {
		while ( !( UCSR0A & (1<<UDRE0)) );	// Wait for empty buffer
		UDR0 = *data++;						// Put data into buffer, sends the data 
	}
}

/**
 * UART transmission function for a single byte
 */
void uart0_send_byte(uint8_t data) {
	while ( !( UCSR0A & (1<<UDRE0)) );	// Wait for empty buffer
 	UDR0 = data;							// Put data into buffer, sends the data 
}


void adc_set_channel(uint8_t chan) {
	ADMUX = (ADMUX & 0b11111000) | chan;
}

/**
 * Main program
 */
int main(void) {		
	timer2_tick_count = 0;
	adc_index = 0;
	
	init_mcu();
	init_uart0_tx();		// Initialize UART TX
	init_timer2();
	init_adc();
	
	sei();
	
	// Main program loop
	while(1) {
		if (timer2_tick_count > TIMER_LATENCY_THRESHOLD) {
			uart0_send_byte(adc_data[0]);
			uart0_send_byte(adc_data[1]);
			uart0_send_byte(adc_data[2]);
			timer2_tick_count = 0;
			
			adc_flag = TRUE;
			adc_index = 0;
			adc_set_channel(adc_index);
			ADCSRA |= _BV(ADSC) | _BV(ADIF);
		}
		
		if (adc_flag && (ADCSRA & _BV(ADIF))) {
			adc_data[adc_index++] = ADCH;
			if (adc_index == ADC_DATA_LENGTH) {
				adc_flag = FALSE;
			}
			else {
				adc_set_channel(adc_index);
				
				ADCSRA |= _BV(ADSC) | _BV(ADIF);
			}
		}
	}
}

/**
 * Initialize UART RX
 */
void init_uart0_tx(void){
	UBRR0L = 12;		// 38400 Baud Rate for 8 MHz Clock
	//UBRR0L = 25;		// 19200 Baud Rate for 8 MHz Clock

	UCSR0A = 0b00000000;// 7 - UART RX complete flag
						// 6 - UART TX complete flag
						// 5 - UART DRE flag
						// 4 - Frame error flag
						// 3 - Data overrun flag
						// 2 - Parity error flag
						// 1 - Double UART speed mode
						// 0 - Multi processor communication mode

	UCSR0B = 0b00001000; // 7 - RX Complete Interrupt Enable
						// 6 - TX Complete Interrupt Enable
						// 5 - UART DRE Interrupt Enable
						// 4 - Enable receiver
						// 3 - Enable transmitter
						// 2 - Character size
						// 1:0 - Ninth data bit

	UCSR0C = 0b00000110;// 7:6 - UART Mode Select
						//		00 - Asynchronous UART
						//		01 - Synchronous UART
						//		10 - Reserved
						//		11 - Master SPI
						// 5:4 - Parity mode
						//		00 - Disabled
						//		01 - Reserved
						//		10 - Enabled, Even
						//		11 - Enabled, Odd
						// 3 - Select stop bit length
						//		0 - 1-bit
						//		1 - 2-bit
						// 2:1 - Character size, with UCSRB.2
						//		000 - 5 bit
						//		001 - 6 bit
						//		010 - 7 bit
						//		011 - 8 bit
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Reserved
						//		111 - 9 bit
						// 0 - Select clock polarity
						//		0 - Rising XCK edge
						//		1 - Falling XCK edge 						
}

void init_timer2(void) {
	TCCR2A = 0b00000010;	// 7:6  COM2A1:0: Compare Match Output A Mode
							// 5:4  COM2B1:0: Compare Match Output B Mode
							// 3:2  Reserved Bits
							// 1:0  WGM21:0: Waveform Generation Mode
							
	TCCR2B = 0b00000111;	// 7  FOC2A: Force Output Compare A
							// 6  FOC2B: Force Output Compare B
							// 5:4  Reserved Bits
							// 3  WGM22: Waveform Generation Mode
							// 2:0  CS22:0: Clock Select
							//	000 - No clk
							//	001 - 1
							//	010 - 8
							//	011 - 32
							//	100 - 64
							//	101 - 128
							//	110 - 256
							//	111 - 1024
	
	ASSR = 0b00000000;		// 6  EXCLK: Enable External Clock Input
							// 5  AS2: Asynchronous Timer/Counter2
							// 4  TCN2UB: Timer/Counter2 Update Busy
							// 3  OCR2AUB: Output Compare Register2 Update Busy
							// 2  OCR2BUB: Output Compare Register2 Update Busy
							// 1  TCR2AUB: Timer/Counter Control Register2 Update Busy
							// 0  TCR2BUB: Timer/Counter Control Register2 Update Busy
	
	TIMSK2 = 0b00000010;	// 7:3 - Reserved Bits
							// 2 - OCIE2B: Timer/Counter2 Output Compare Match B Interrupt Enable
							// 1  OCIE2A: Timer/Counter2 Output Compare Match A Interrupt Enable
							// 0  TOIE2: Timer/Counter2 Overflow Interrupt Enable
							
	OCR2A = 250;						

}

void init_adc(void) {
	ADMUX = 0b01100000;		// 7:6  REFS1:0: Reference Selection Bits
							// 5  ADLAR: ADC Left Adjust Result
							// 4:0  MUX4:0: Analog Channel and Gain Selection Bits
							
	ADCSRA = 0b10000101;	// 7  ADEN: ADC Enable
							// 6  ADSC: ADC Start Conversion
							// 5  ADATE: ADC Auto Trigger Enable
							// 4  ADIF: ADC Interrupt Flag
							// 3  ADIE: ADC Interrupt Enable
							// 2:0  ADPS2:0: ADC Prescaler Select Bits
							
	ADCSRB = 0b00000000;	// 7, 5:3 - Reserved Bits
							// 2:0  ADTS2:0: ADC Auto Trigger Source

	DIDR0 = 0b00000111;		// 7:0 - ADC7:0 Digital Input Disable
							//	=> Disable digital input buffer to reduce power consumption 
}

void init_mcu(void) {
	SMCR = 0b00000000;	// 7:4 - Reserved bits
						// 3:1 - Sleep Mode
						//		000 - Idle
						//		001 - ADC Noise Reduction
						//		010 - Power-down
						//		011 - Power-save
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Standby
						//		111 - Extended Standby
						// 0 - Sleep Enable
						
	MCUCR |= (0 << 6);	// 6 - BOD Sleep
	MCUCR |= (0 << 5);	// 5 - BOD Sleep Enable
	
	PRR = 0b10111100;	// 7 - Power Reduction TWI
						// 6 - Power Reduction Timer2
						// 5 - Power Reduction Timer0
						// 4 - Power Reduction UART1
						// 3 - Power Reduction Timer1
						// 2 - Power Reduction SPI
						// 1 - Power Reduction UART0
						// 0 - Power Reduction ADC
}
